%% tclpext__type_graph.pl module to display graphs of the type hierarchy
%% Copyright (C) 2001,2002 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                   TCLP                                    %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/beaune/contraintes/coquery/cvs/contraintes/tclpchr/src/tclpext__type_graph.pl,v 1.1 2002/09/04 15:05:57 coquery Exp $

:- module(tclpext__type_graph, [write_type_graph/2,
                                save_type_graph/1,
                                save_gif_type_graph/1,
                                save_ps_type_graph/1
                                ]).

:- use_module(tclp__arity).
:- use_module(tclp__type_order).
:- use_module(library(ugraphs)).
:- use_module(library(lists)).
:- use_module(prolog_dot).
:- use_module(library(charsio)).
:- use_module(library(system)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Building the graph of direct type relations
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% First we build the set of pairs tau1-tau2 such that
%% 1) tau1 < tau2
%% 2) for all types tau3,
%%    tau3<>tau1 & tau3<>tau2 ==> not tau1 < tau3 & tau3 < tau2
make_direct_linked_pairs(Pairs) :-
        findall(Tau2-Tau1,
                (
                  chr:find_constraint(lt(Tau1,Tau2),_),
                  \+ (
                       chr:find_constraint(lt(Tau1,X),_),
                       chr:find_constraint(lt(X,Tau2),_)
                     )
                ),
                Pairs).

%% Unifies the arguments of two type constructor, according to the
%% iota function that maps argument between type constructors
unify_args(Type1,Type2) :-
        functor(Type1,C1,N1), functor(Type1b,C1,N1),
        functor(Type2,C2,N2), functor(Type2b,C2,N2),
        unifyToArity(Type1b), unifyToArity(Type2b),
        copy_term(Type1b-Type2b, Type1-Type2).

%% apply unify_args to each pair of the argument
unify_args_all([]).
unify_args_all([X-Y|Rest]) :-
        unify_args(X,Y),!,
        unify_args_all(Rest).

%% build_pairs_with_args(Types, InitPairs, FinalPairs)
%% Types is a list of type constructors with variables as their arguments
%% InitPairs is a list of pairs of atom corresponding to elements of Types
%% FinalPairs is similar to InitPairs except that the atoms are replaced by
%% their conterpart in Types.
build_pairs_with_args(_,[],[]).
build_pairs_with_args(Types, [A-B|Rest1], [AArgs-BArgs|Rest2]) :-
        once((member(AArgs, Types), functor(AArgs, A, _))),
        once((member(BArgs, Types), functor(BArgs, B, _))),!,
        build_pairs_with_args(Types,Rest1,Rest2).
build_pairs_with_args(Types, [_|Rest1], Rest2) :-
        build_pairs_with_args(Types,Rest1,Rest2).
        

%% returns a list of types with their argument
get_all_types(Types) :-
        chr:findall_constraints(arity(_,_), Constraints),
        make_type_list(Constraints,Types).
make_type_list([],[]).
make_type_list([ arity(bottom,0)#_ | RestC], RestT) :-
        make_type_list(RestC, RestT).
make_type_list([ arity(top,0)#_ | RestC], RestT) :-
        make_type_list(RestC, RestT).
make_type_list([ arity(Cons,N)#_ | RestC], [Type|RestT]) :-
        functor(Type,Cons,N),
        make_type_list(RestC, RestT).

%% Convert the edges to atoms
convert_edges([],[]).
convert_edges([A-B|Rest1],[AAtom-BAtom|Rest2]) :-
        write_to_chars(A,AChars), atom_chars(AAtom,AChars),
        write_to_chars(B,BChars), atom_chars(BAtom,BChars),
        convert_edges(Rest1,Rest2).

%% Convert types to atoms
convert_types([],[]).
convert_types([A|As],[B|Bs]) :-
        write_to_chars(A,Chars), atom_chars(B,Chars),
        convert_types(As,Bs).
        

%% build a graphs that represents the type structure
build_type_graph(Graph) :-
        init_order,!,
        get_all_types(Types),
        make_direct_linked_pairs(SimplePairs),
        build_pairs_with_args(Types,SimplePairs,Edges),
        unify_args_all(Edges),
        numbervars(Types,0,_),
        convert_types(Types,TypesAtom),
        convert_edges(Edges,EdgesAtom),
        vertices_edges_to_ugraph(TypesAtom, EdgesAtom, Graph),!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Output the graph to a .dot file using prolog_dot
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

write_type_graph(Stream, Title) :-
        build_type_graph(Graph),
        write_atom_graph(Stream,Title,Graph).

write_color_type_graph(Stream, Title) :-
        build_type_graph(Graph),
        write_color_atom_graph(Stream,Title,Graph).

save_type_graph(Filename) :-
        open(Filename, write, Stream),
        write_type_graph(Stream, '"Types for TCLP"'),
        close(Stream).

%% FIXME use a pipe instead of a file for the .dot part
save_gif_type_graph(Filename) :-
        once((atom_concat(Prefix,'.gif',Filename) ; Prefix=Filename)),
        atom_concat(Prefix , '.dot', DotFilename),
        save_type_graph(DotFilename),
        format_to_chars('dot ~a -o ~a -Tgif',[DotFilename,Filename],Chars),
        atom_chars(Cmd,Chars),
        shell(Cmd).

save_ps_type_graph(Filename) :-
        once((atom_concat(Prefix,'.ps',Filename) ; Prefix=Filename)),
        atom_concat(Prefix , '.dot', DotFilename),
        save_type_graph(DotFilename),
        format_to_chars('dot ~a -o ~a -Tps',[DotFilename,Filename],Chars),
        atom_chars(Cmd,Chars),
        shell(Cmd).

